
// Operations on cyclic distributed ranges.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"

inline int max(int x, int y) {return x > y ? x : y ; }


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CyclicRangeRep': cyclic-distributed template range.          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* CyclicRangeRep :: subrng(const int _extent,
                                   const int _base, const int _stride) {
  return new CyclicSubRangeRep(_extent, _base, _stride, this,
                               kernel, numLayers) ;
}

Format CyclicRangeRep :: format() const {return DIST_CYCLIC ; }

int CyclicRangeRep :: ker_glb(const int glb, int* sub) const {

  *sub = glb / numProcs ;
  return glb % numProcs ;
}

int CyclicRangeRep :: block(const int ker_glb, int* sub, int* count) const {

  // `ker_glb' is a global subscript in the kernel range.

  *sub   = 0 ;
  *count = numLayers ;

  return ker_glb ;
}

int CyclicRangeRep :: block_step(int* sub_stp) const {

  *sub_stp = 1 ;
  return numProcs ;
}

int CyclicRangeRep :: width() const {return numLayers ; }

StrideRep* CyclicRangeRep :: vector(StrideRep* _kernel,
                                    const int _length) const {
  return new StrideRep(_kernel, _length) ;
}

int CyclicRangeRep :: glb(const int sub, const int ker_glb) const {
  return ker_glb + numProcs * sub ;
}

void CyclicRangeRep :: show() {
  AdlibPrintf("CyclicRangeRep :: show: size = %d ker = %p lev = %d dim = %p\n",
              extent, kernel, level, (DimensionRep*) dimension) ; 
}

MEMORY_POOL_DEFN(CyclicRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CyclicSubRangeRep': cyclic-distributed subrange.             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* CyclicSubRangeRep :: subrng(const int _extent,
                                      const int _base, const int _stride) {
  return new CyclicSubRangeRep(_extent,
                               base + stride * _base, stride * _stride,
                               parent, kernel, numLayers) ;
}

RangeRep* CyclicSubRangeRep :: tem() {return parent ; }

int CyclicSubRangeRep :: bas() const {return base ; }

int CyclicSubRangeRep :: str() const {return stride ; }

Format CyclicSubRangeRep :: format() const {return DIST_CYCLIC ; }

int CyclicSubRangeRep :: ker_glb(const int glb, int* sub) const {
  const int tem = base + stride * glb ;

  *sub = tem / numProcs ;
  return tem % numProcs ;
}

void CyclicSubRangeRep :: euler() {

  // Compute `f', the GCD of `numProcs' and `stride', and `f_s',
  // a solution of
  //
  //   f_s * s = f  mod  p
  // 
  // (s = stride, p = numProcs)

  int g, g_s ;

  f   = numProcs ;
  f_s = 0 ;
  if(stride > 0) {
    g   = stride ;
    g_s = 1 ;
  }
  else {
    g   = -stride ;
    g_s = -1 ;
  }
  while(g) {
    const int r = f / g ;

    const int temp1 = g ;
    g    = f - r * g ;
    f    = temp1 ;
  
    const int temp2 = g_s ;
    g_s  = f_s - r * g_s ;
    f_s  = temp2 ;
  }

  k_step = numProcs / f ;
}

int CyclicSubRangeRep :: block(const int ker_glb, int* sub, int* count) const {

  // `ker_glb' is a global subscript in the kernel range.

  // Define `k_lo' and `k_hi' for current value of `ker_glb'

  // Compute the smallest and largest solutions of
  //
  //     base + k * s = crd    mod  p
  //
  // in the range 0, ..., extent - 1, viz
  //
  //     k_lo = a + b * ceiling(- a / b)
  //     k_hi = a + b * floor((extent - 1 - a) / b)
  //
  // where a = c / f * f_s, b = p / f, c = crd - base.
  //
  // Solutions exist iff `f' divides `c'.

  const int disp = ker_glb - base ;
  if(disp % f == 0) {
    int k_lo, k_hi ;

    const int a = (disp / f) * f_s ;
    const int b = k_step ;

    if(a >= 0)
      k_lo = a - b * (a / b) ;
    else
      k_lo = a + b * ((-a + b - 1) / b) ;

    if(extent > a)
      k_hi = a + b * ((extent - 1 - a) / b) ;
    else
      k_hi = a - b * ((-extent + a + b) / b) ;

    *sub   = (base + stride * k_lo) / numProcs ;
    *count = max((k_hi - k_lo) / k_step + 1, 0) ;

    return k_lo ;
  }
  else {
    *sub   = 0 ;
    *count = 0 ;

    return 0 ;
  }
}

int CyclicSubRangeRep :: block_step(int* sub_stp) const {

  //*sub_stp = (stride * k_step) / numProcs ;
  *sub_stp = stride / f ;
  return k_step ;
}

RangeRep* CyclicSubRangeRep :: subker(int* ker_bas, int* ker_str) const {
  *ker_bas = base % f ;
  *ker_str = f ;

  return kernel->subrng(kernel->size() / f, *ker_bas, *ker_str) ;
}

int CyclicSubRangeRep :: width() const {return numLayers ; }

StrideRep* CyclicSubRangeRep :: vector(StrideRep* _kernel,
                                       const int _length) const {
  return new StrideRep(_kernel, _length) ;
}

int CyclicSubRangeRep :: glb(const int sub, const int ker_glb) const {
  return (ker_glb + numProcs * sub - base) / stride ;
}

void CyclicSubRangeRep :: show() {
  AdlibPrintf("CyclicSubRangeRep :: show: size = %d ker = %p lev = %d dim = %p\n",
              extent, kernel, level, (DimensionRep*) dimension) ; 
}

MEMORY_POOL_DEFN(CyclicSubRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `ADAPTCyclicRangeRep', `ADAPTCyclicSubRangeRep':            //
//     Cyclic-distributed ranges yielding arrays with ADAPT layout.    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* ADAPTCyclicRangeRep :: subrng(const int _extent,
                                        const int _base, const int _stride) {
  return new ADAPTCyclicSubRangeRep(_extent, _base, _stride,
                                    this, kernel, numLayers) ;
}


RangeRep* ADAPTCyclicSubRangeRep :: subrng(const int _extent,
                                           const int _base, const int _stride) {
  return new ADAPTCyclicSubRangeRep(_extent,
				    base + stride * _base, stride * _stride,
                                    parent, kernel, numLayers) ;
}

int ADAPTCyclicSubRangeRep :: width() const {
  const int top = stride > 0 ? base + stride * (extent - 1) + 1 : base + 1 ;
  return (top + numProcs - 1) / numProcs ;
}


// ------------------------------------------------------------------
// Things to do
//

